#!/usr/bin/env python
# -*- coding: utf-8 -*-

# A demo to interactively showcase all the features.
# Copyright Gummbum, 4/18/2017. Free software, no warranty express or implied.

# Using the demo.
#
# Hopefully this is intuitive. When the program starts the title bar displays "Press a key to start". Then use the
# number and letter keys displayed in the HUD to manipulate the HUD settings.
#
# To change the FONT SIZE, scroll the middle mouse wheel. To reset it, press the 2 key.

# You don't have to subclass HUD, but I did to make the code cleaner, and benefit from IDE inspection. It really helps
# in this demo because this is a very busy, complex example.
#
# Outline looks best with a fat font. Gumm likes OColor=Yellow. Especially attractive with GColor in (deepskyblue,
# MediumPurple2, Firebrick4). Almost any FgColor looks stunning with those combos.

import pygame
from pygame.locals import *

import hudlight


class MyHUD(hudlight.HUD):
    """
    This class does:
    - adds the HUD items
    - provides methods to update each HUD item (calling code looks neater)
    - a monolithic update_all() method
    """

    def __init__(self, fontname=None, fontsize=None, sysfontname=None, bold=False, italic=False, underline=False,
                 color=None, background=None, gcolor=None, ocolor=None, scolor=None,
                 antialias=True, alpha=1.0):
        super(MyHUD, self).__init__(fontname, fontsize, sysfontname, bold, italic, underline,
                                    color, background, gcolor, ocolor, scolor, antialias=antialias, alpha=alpha)
        self.add('font', '1=Font ({}), 2=Size ({})', fontname or sysfontname, fontsize)
        self.add('style', '3=Bold ({}), 4=Italic ({}) 5=Under ({})', bold, italic, underline)
        self.add('antialias', '6=AntiAlias ({})', self._antialias)
        self.add('color', '7=Color ({}) g=GColor ({})', color, gcolor)
        self.add('background', '8=Background ({})', background)
        self.add('shadow', 's=SColor ({})', self._scolor)
        self.add('outline', 'o=OColor ({})', self._ocolor)
        self.add('fps', 'FPS {:0.0f}', 0)
        self.add('mouse', 'Mouse {0[0]}, {0[1]}', (0, 0))
        self.add('angle', 'Angle {:0.0f}', 0.0)
        self.add('perf', 'Rendered/sec {}', 0)
        hudlight.set_font_template("fonts/%s")

    def update_all(self, fontname, fontsize, bold, italic, underline, antialias, color, background, gcolor,
                   shadow_color, outline_color, fps, mouse_pos, angle):
        self.update_font(fontname, fontsize)
        self.update_style(bold, italic, underline)
        self.update_antialias(antialias)
        self.update_color(color, gcolor)
        self.update_background(background)
        self.update_shadow(shadow_color)
        self.update_outline(outline_color)
        self.update_fps(fps)
        self.update_mouse(mouse_pos)
        self.update_angle(angle)

    def update_font(self, font, size):
        self.update_item('font', font, size)

    def update_style(self, bold, italic, underline):
        self.update_item('style', bold, italic, underline)

    def update_antialias(self, antialias):
        self.update_item('antialias', antialias)

    def update_color(self, fgcolor, gcolor):
        self.update_item('color', fgcolor, gcolor)

    def update_background(self, bgcolor):
        self.update_item('background', bgcolor)

    def update_shadow(self, scolor):
        self.update_item('shadow', scolor)

    def update_outline(self, ocolor):
        self.update_item('outline', ocolor)

    def update_fps(self, fps):
        self.update_item('fps', fps)

    def update_mouse(self, mouse_pos):
        self.update_item('mouse', mouse_pos)

    def update_angle(self, angle):
        self.update_item('angle', angle)

    def update_perf(self):
        self.update_item('perf', int(self.fps()))


class MyUI(object):
    """
    This class:
    - contains runtime settings that the user can change via key presses
    - creates a keymap: pygame keys -> cycle_*() methods (calling code is much neater; see do_key())
    - circular linked lists for the menus to cycle through (see the circular linked lists at the end of
      __init__ if you wish to customize the menu data; this is pretty easy, but read the notes)
    - cycle_*() methods (do the work of cycling a menu choice via its circular linked list)
    """

    def __init__(self, fontname, fontsize, bold, italic, underline, antialias, color, background, gcolor, shadow, outline,
                 keys=None):
        # Current values.
        self.fontname = fontname
        self.fontsize = fontsize
        self.bold = bold
        self.italic = italic
        self.underline = underline
        self.antialias = antialias
        self.color = color
        self.background = background
        self.gcolor = gcolor
        self.scolor = shadow
        self.ocolor = outline
        self.shadow = 1, 1

        # Keymap: pygame key -> cycle_*() method. do_key() handles the lookups and calls.
        self.keymap = {}
        if keys:
            self.set_keymap(keys)
        self.is_interactive = False

        # Circular linked lists, used by the cycle*() methods.
        # Usage: new_font = next_font.get(current_font, 'default').
        # Customization: see make_cll() docstring for explanation of args.
        # NOTE: make sure there is a default in position 0, even if it repeats position 1. All the code
        # relies on this.
        self.cll_fontname = self._make_cll(
            'VeraBd.ttf', 'VeraBd.ttf', 'Vera.ttf', 'Boogaloo.ttf', 'Bubblegum_Sans.ttf', 'CherryCreamSoda.ttf',
            'Roboto_Condensed.ttf', 'sans', 'times', printme=False)
        self.cll_fontsize = self._make_cll(28, 28, printme=False)
        self.cll_bold = self._make_cll(False, True, False, printme=False)
        self.cll_italic = self._make_cll(False, True, False, printme=False)
        self.cll_underline = self._make_cll(False, True, False, printme=False)
        self.cll_antialias = self._make_cll(True, True, False, printme=False)
        self.cll_color = self._make_cll(*'red2 red2 orange2 yellow2 green3 blue2'.split(), printme=False)
        self.cll_background = self._make_cll(None, None, 'midnightblue', 'skyblue', printme=False)
        self.cll_gcolor = self._make_cll(
            None, None, *'deepskyblue gold cornflowerblue mediumpurple2 firebrick4'.split(), printme=False)
        self.cll_scolor = self._make_cll('black', None, *'black grey20 grey40 darkviolet yellow'.split(), printme=False)
        self.cll_ocolor = self._make_cll(None, None, *'black grey25 grey40 darkviolet yellow'.split(), printme=False)

    @staticmethod
    def _make_cll(*words, **kwargs):
        """Make a circular linked list from arbitrary words

        :param words: contents are: default, choice1, choice2, ...
        :param kwargs: printme=True for a debug print
        :return: None
        """
        printme = kwargs.get('printme', False)
        cll = {}
        null = '__null__'
        init = first = key = null
        for word in words:
            if init is null:
                cll['default'] = word
                init = word
            elif first is null:
                key = first = word
            else:
                cll[key] = word
                key = word
        cll[key] = first
        if printme:
            print([(k, cll[k]) for k in (['default'] + list(words))])
        return cll

    def set_keymap(self, keys):
        # Map the event keys to cycle*() methods.
        names = 'fontname fontsize bold italic underline antialias color background gcolor scolor ocolor'.split()
        methods = [getattr(self, 'cycle_' + n) for n in names]
        self.keymap = dict(zip(keys, methods))

    def do_key(self, key, hud):
        # Look up the key and call its assigned cycle*() method
        if key in self.keymap:
            self.keymap[key](hud)

    def set_interactive(self, hud):
        # Call once to go from initial splash to interactive mode
        if not self.is_interactive:
            self._reset_defaults()
            self._set_font(hud)
            hud.set_color(self.color)
            hud.set_background(self.background)
            hud.set_gcolor(self.gcolor)
            hud.set_antialias(self.antialias)
            hud.set_scolor(self.scolor)
            hud.set_ocolor(self.ocolor)
            hud.set_shadow(self.shadow)
            self.is_interactive = True

    def inc_fontsize(self, hud):
        self.fontsize += 1
        hud.set_fontsize(self.fontsize)

    def dec_fontsize(self, hud):
        if self.fontsize <= 5:
            return
        self.fontsize -= 1
        hud.set_fontsize(self.fontsize)

    def cycle_fontname(self, hud=None):
        self.fontname = self._cycle(self.cll_fontname, self.fontname)
        if hud:
            if 'ttf' in self.fontname:
                hud.set_fontname(self.fontname)
            else:
                hud.set_sysfontname(self.fontname)
        return self.fontname

    def cycle_fontsize(self, hud=None):
        self.fontsize = self._cycle(self.cll_fontsize, self.fontsize)
        hud and hud.set_fontsize(self.fontsize)
        return self.fontsize

    def cycle_bold(self, hud=None):
        self.bold = self._cycle(self.cll_bold, self.bold)
        hud and hud.set_bold(self.bold)
        return self.bold

    def cycle_italic(self, hud=None):
        self.italic = self._cycle(self.cll_italic, self.italic)
        hud and hud.set_italic(self.italic)
        return self.italic

    def cycle_underline(self, hud=None):
        self.underline = self._cycle(self.cll_underline, self.underline)
        hud and hud.set_underline(self.underline)
        return self.underline

    def cycle_antialias(self, hud):
        self.antialias = self._cycle(self.cll_antialias, self.antialias)
        hud.set_antialias(self.antialias)
        return self.antialias

    def cycle_color(self, hud):
        self.color = self._cycle(self.cll_color, self.color)
        hud.set_color(self.color)
        return self.color

    def cycle_background(self, hud):
        self.background = self._cycle(self.cll_background, self.background)
        hud.set_background(self.background)
        return self.background

    def cycle_gcolor(self, hud):
        self.gcolor = self._cycle(self.cll_gcolor, self.gcolor)
        hud.set_gcolor(self.gcolor)
        return self.gcolor

    def cycle_scolor(self, hud):
        self.scolor = self._cycle(self.cll_scolor, self.scolor)
        hud.set_scolor(self.scolor)
        return self.scolor

    def cycle_ocolor(self, hud):
        self.ocolor = self._cycle(self.cll_ocolor, self.ocolor)
        hud.set_ocolor(self.ocolor)
        return self.ocolor

    def _reset_defaults(self):
        # Set everything to ccl['default'] value
        for name in 'fontname fontsize bold italic underline antialias color background gcolor scolor ocolor'.split():
            cll = getattr(self, 'cll_' + name)
            setattr(self, name, cll['default'])

    @staticmethod
    def _cycle(cll, value):
        return cll.get(value, cll['default'])

    def _set_font(self, hud):
        if 'ttf' in self.fontname:
            # Font
            hud.set_font(self.fontname, self.fontsize, None, self.bold, self.italic, self.underline)
        else:
            # SysFont
            hud.set_font(None, self.fontsize, self.fontname, self.bold, self.italic, self.underline)


def run_demo():
    import math

    # The basic pygame things
    pygame.init()
    clock = pygame.time.Clock()
    screen = pygame.display.set_mode((1024, 768))
    screen_rect = screen.get_rect()
    pygame.display.set_caption('HUD Light - Demo 3, Full Interactive | Welcome. Press a key to start | ESC or Close to quit')
    clear_color = Color('grey20')

    # Some runtime things for the HUD to watch
    origin = screen_rect.center
    radius = screen_rect.centerx * 4 / 5
    angle = 0
    mouse_pos = 0, 0
    max_fps = 60

    # Some decorative lines to break up the monotone background
    def make_lines(lines):
        radius = max(screen_rect.w // 2, screen_rect.h // 2)
        for angle in range(0, 360, 2):
            radians_ = math.radians(angle)
            x = origin[0] + radius * math.cos(radians_)
            y = origin[1] + radius * math.sin(radians_)
            lines.append((origin, (x, y)))

    decorative_lines = []
    make_lines(decorative_lines)

    # Create the user interface. This is driven by key presses, and configures the HUD on the fly.
    # Keymaps: ui.do_key() -> ui.cycle_*()
    ui = MyUI(
        'fixed', 26, True, True, True, True,  # fontname, fontsize, bold, italic, underline, antialias
        'hotpink4', None, None,  # color, background, gcolor
        'hotpink', None,  # scolor, ocolor
        # keys, see MyUI.set_keymap()
        (K_1, K_2, K_3, K_4, K_5, K_6,  # fontname, fontsize, bold, italic, underline, antialias
         K_7, K_8,        # color, background
         K_g, K_s, K_o))  # gcolor, scolor, ocolor

    # Make our HUD
    hud = MyHUD(None, ui.fontsize, ui.fontname, ui.bold, ui.italic, ui.underline,
                ui.color, ui.background, ui.gcolor, ui.ocolor, ui.scolor)

    running = True
    ticks = 0

    while running:
        clock.tick(max_fps)
        ticks += 1

        for e in pygame.event.get():
            if e.type == MOUSEMOTION:
                mouse_pos = e.pos
            elif e.type == KEYDOWN:
                if e.key in (K_ESCAPE, K_q):
                    running = False
                elif not ui.is_interactive:
                    ui.set_interactive(hud)
                    pygame.display.set_caption('HUD Light - Demo 3, Full Interactive | See HUD on screen for keys | Q or ESC to quit')
                else:
                    ui.do_key(e.key, hud)
            elif e.type == MOUSEBUTTONDOWN:
                if ui.is_interactive:
                    if e.button == 4:
                        ui.inc_fontsize(hud)
                    elif e.button == 5:
                        ui.dec_fontsize(hud)
            elif e.type == QUIT:
                running = False

        # Update the spinner
        x = math.cos(angle) * radius + origin[0]
        y = math.sin(angle) * radius + origin[1]
        angle = (angle + 0.005) % (2 * math.pi)

        # Update the HUD values
        if ticks % 2 == 0:
            hud.update_all(
                ui.fontname, ui.fontsize, ui.bold, ui.italic, ui.underline, ui.antialias,
                ui.color, ui.background, ui.gcolor,
                ui.scolor, ui.ocolor,
                clock.get_fps(), mouse_pos, math.degrees(angle))
        if ticks % max_fps == 0:
            hud.update_perf()

        # Draw
        screen.fill(clear_color)
        # background
        for p1, p2 in decorative_lines:
            pygame.draw.line(screen, Color('purple'), p1, p2)
        # spinner
        pygame.draw.line(screen, Color('white'), origin, (x, y))
        # hud
        hud.draw(screen)
        pygame.display.flip()


if __name__ == '__main__':
    run_demo()
